from ast import Expression
from collections import deque
from dataclasses import dataclass
from enum import Enum
import re
from typing import Dict, List, Tuple, Union
import operator
import numpy as np
from scipy import stats
import random
from fractions import Fraction
from lib.typing import *

KEYWORDS = ["fn", "int", "real", "while", "if", "else"]
ASSIGNMENT_OPERATORS = ["~", "-=", "\+="]
OPERATORS = [">=", ">", "==", "\|\|", "<=", "<", "&&", "\+", "\*", "-"]

PATTERNS = [
    ("OPEN_BRACKET", re.compile("\(")),
    ("CLOSE_BRACKET", re.compile("\)")),
    ("OPEN_CURLY", re.compile("\{")),
    ("CLOSE_CURLY", re.compile("\}")),
    ("COMMA", re.compile(",")),
    ("COLON", re.compile(":")),
    ("SEMICOLON", re.compile(";")),
    ("REAL_LITERAL", re.compile("(-)?\d+\.\d+")),
    ("INT_LITERAL", re.compile("(-)?\d+")),
    ("KEYWORD", "|".join(KEYWORDS)),
    ("IDENTIFIER", re.compile("[a-zA-Z][a-zA-Z0-9]*")),
    ("ASSIGNMENT_OPERATOR", re.compile("|".join(ASSIGNMENT_OPERATORS))),
    ("OPERATOR", re.compile("|".join(OPERATORS))),
    ("ASSIGNMENT_OPERATOR", re.compile("=")),
    ("COMMENT", re.compile("\/\/.*\n")),
    ("BLOCK_COMMENT", re.compile("\/\*(\*(?!\/)|[^*])*\*\/"))
]

DISTRIBUTIONS = {
    "Bernoulli": ([PrimType.REAL], PrimType.INT),
    "Uniform": ([PrimType.REAL, PrimType.REAL], PrimType.REAL),
    "Gaussian": ([PrimType.REAL, PrimType.REAL], PrimType.REAL)
}

class Lexer:
    def __init__(self, expr):
        self.expr = expr
        self.tokens = []

    def matches(self, re_expr):
        match = re.match(re_expr, self.expr)
        if match != None:
            str = match[0]
            self.expr = self.expr[len(str):]
            return str
        return None

    def process(self):
        self.expr = self.expr.lstrip()
        while len(self.expr) != 0:
            found = False
            for (name, re_expr) in PATTERNS:
                if str := self.matches(re_expr):
                    self.tokens.append((name, str))
                    found = True
                    break
            if not found:
                print(self.expr)
                self.error()
            self.expr = self.expr.lstrip()

    def error(self):
        raise Exception("Lexer Error")

OPERATOR_TABLE = {
    "*": 100,
    "+": 90,
    "-": 90,
    "==": 80,
    "<": 80,
    ">": 80,
    "<=": 80,
    ">=": 80,
    "&&": 70,
    "||": 60
}

class Parser:
    def __init__(self, tokens):
        tokens = [token for token in tokens if token[0] not in ["COMMENT", "BLOCK_COMMENT"]]
        self.tokens = deque(tokens)

    def advance(self):
        if len(self.tokens) == 0:
            self.error()
        return self.tokens.popleft()

    def peek(self):
        if len(self.tokens) == 0:
            self.error()
        return self.tokens[0][0]
    
    def peek_value(self):
        if len(self.tokens) == 0:
            self.error()
        return self.tokens[0][1]

    def advance_type(self, type):
        if len(self.tokens) == 0:
            self.error()
        token = self.tokens.popleft()
        if token[0] != type:
            self.error()
        return token[1]

    def advance_keyword(self, keyword):
        if len(self.tokens) == 0:
            self.error()
        token = self.tokens.popleft()
        if token[0] != "KEYWORD" or token[1] != keyword:
            self.error()

    def advance_data_type(self):
        data_type = self.advance_type("KEYWORD")
        mapping = {
            "int": PrimType.INT,
            "real": PrimType.REAL
        }
        if not data_type in mapping:
            self.error()
        return mapping[data_type]

    def parse_arg_list(self):
        args = []
        self.advance_type("OPEN_BRACKET")
        while True:
            args.append(self.parse_expression(0))
            if self.peek() != "COMMA":
                break
            self.advance_type("COMMA")
        self.advance_type("CLOSE_BRACKET")
        return args

    def parse_atom(self):
        match self.advance():
            case ("REAL_LITERAL", value):
                return RealNode(Fraction(value))
            case ("INT_LITERAL", value):
                return IntNode(int(value))
            case ("IDENTIFIER", name):
                if self.peek() != "OPEN_BRACKET":
                    return IdentifierNode(name)
                else:
                    return DistributionNode(operator=name, args=self.parse_arg_list())
            case ("OPEN_BRACKET", _):
                expr = self.parse_expression(0)
                self.advance_type("CLOSE_BRACKET")
                return expr

    def parse_expression(self, min_prec):
        result = self.parse_atom()
        while self.peek() == "OPERATOR" and OPERATOR_TABLE[self.peek_value()] >= min_prec:
            op = self.advance_type("OPERATOR")
            prec = OPERATOR_TABLE[op]
            next_min_prec = prec + 1
            rhs = self.parse_expression(next_min_prec)
            result = BinaryOperator(op, result, rhs)
        return result

    def parse_assignment(self):
        name = self.advance_type("IDENTIFIER")
        op = self.advance_type("ASSIGNMENT_OPERATOR")
        expr = self.parse_expression(0)
        self.advance_type("SEMICOLON")
        if op == "~":
            return AssignmentStatement(
                name=name,
                expr=UnaryOperator(
                    operator="sample",
                    operand=expr
                )
            )
        elif op == "-=":
            return AssignmentStatement(
                name=name,
                expr=BinaryOperator(
                    operator="-", 
                    operand1=IdentifierNode(name), 
                    operand2=expr
                )
            )
        elif op == "+=":
            return AssignmentStatement(
                name=name,
                expr=BinaryOperator(
                    operator="+", 
                    operand1=IdentifierNode(name), 
                    operand2=expr
                )
            )
        elif op == "=":
            return AssignmentStatement(
                name=name,
                expr=expr
            )

    def parse_if(self):
        self.advance_type("OPEN_BRACKET")
        cond_first = self.parse_expression(0)
        self.advance_type("CLOSE_BRACKET")
        stmt_first = self.parse_block()
        curr = [(cond_first, stmt_first)]
        while True:
            if self.peek() == "KEYWORD" and self.peek_value() == "else":
                self.advance_keyword("else")
                if self.peek() == "KEYWORD" and self.peek_value() == "if":
                    self.advance_keyword("if")
                    self.advance_type("OPEN_BRACKET")
                    cond = self.parse_expression(0)
                    self.advance_type("CLOSE_BRACKET")
                    stmt = self.parse_block()
                    curr.append((cond, stmt))
                else:
                    stmt = self.parse_block()
                    curr.append((BoolNode(value=True), stmt))
            else:
                break
        return IfStatement(branches=curr)

    def parse_while(self):
        self.advance_type("OPEN_BRACKET")
        cond = self.parse_expression(0)
        self.advance_type("CLOSE_BRACKET")
        stmt = self.parse_block()
        return WhileStatement(cond, stmt)

    def parse_block(self):
        self.advance_type("OPEN_CURLY")
        statements = []
        while self.peek() != "CLOSE_CURLY":
            if self.peek() == "IDENTIFIER":
                statements.append(self.parse_assignment())
            else:
                keyword = self.advance_type("KEYWORD")
                if keyword == "if":
                    statements.append(self.parse_if())
                elif keyword == "while":
                    statements.append(self.parse_while())
                else:
                    self.error()

        self.advance_type("CLOSE_CURLY")
        return BlockStatement(statements=statements)

    def parse_function(self):
        self.advance_keyword("fn")
        name = self.advance_type("IDENTIFIER")
        self.advance_type("OPEN_BRACKET")
        args = []
        if self.peek() == "IDENTIFIER":
            while True:
                arg_name = self.advance_type("IDENTIFIER")
                self.advance_type("COLON")
                arg_type = self.advance_data_type()
                args.append((arg_name, arg_type))
                if self.peek() != "COMMA":
                    break
                self.advance_type("COMMA")
        self.advance_type("CLOSE_BRACKET")
        stmt = self.parse_block()
        return Function(name=name, args=args, body=stmt)

    def parse(self):
        funcs = []
        while len(self.tokens) != 0:
            funcs.append(self.parse_function())
        return Program(funcs)

    def error(self):
        print(self.tokens)
        raise Exception("Parser Error") 

Env = Dict[str, Type]

class TypeChecker:
    def check_args(self, args) -> Env:
        env = {}
        for (name, arg_type) in args:
            if name in env:
                self.error()
            env[name] = arg_type
        return env

    def check_expr(self, env: Env, expr: Expression) -> Type:
        match expr:
            case IntNode():
                return PrimType.INT
            case RealNode():
                return PrimType.REAL
            case BoolNode():
                return PrimType.BOOL
            case IdentifierNode(name=name):
                if name not in env:
                    self.error()
                return env[name]
            case DistributionNode(operator=operator, args=args):
                if operator not in DISTRIBUTIONS:
                    self.error()
                (spec, result) = DISTRIBUTIONS[operator]
                if len(spec) != len(args):
                    self.error()
                for (arg, arg_type) in zip(args, spec):
                    if self.check_expr(env, arg) != arg_type:
                        self.error()
                return DistType(result)
            case UnaryOperator(operator="sample", operand=operand):
                dist_type = self.check_expr(env, operand)
                match dist_type:
                    case DistType(base=base):
                        return base
                    case _:
                        self.error()
            case BinaryOperator(operator=operator, operand1=operand1, operand2=operand2):
                type1 = self.check_expr(env, operand1)
                type2 = self.check_expr(env, operand2)
                if operator in ["==", "<", ">", "<=", ">="]:
                    if type1 != type2:
                        self.error()
                    return PrimType.BOOL
                elif operator in ["+", "-", "*"]:
                    if type1 != type2:
                        self.error()
                    elif type1 == PrimType.BOOL:
                        self.error()
                    return type1
                elif operator in ["||", "&&"]:
                    if type1 != PrimType.BOOL or type2 != PrimType.BOOL:
                        self.error()
                    return PrimType.BOOL

    def check_statement(self, env: Env, stmt: Statement):
        match stmt:
            case AssignmentStatement(name=name, expr=expr):
                expr_type = self.check_expr(env, expr)
                if name in env:
                    if env[name] != expr_type:
                        self.error()
                else:
                    env[name] = expr_type
            case IfStatement(branches=branches):
                for (cond, block) in branches:
                    if self.check_expr(env, cond) != PrimType.BOOL:
                        self.error()
                    env_copy = env.copy()
                    self.check_block(env_copy, block)
            case WhileStatement(guard=guard, body=body):
                if self.check_expr(env, guard) != PrimType.BOOL:
                    self.error()
                env_copy = env.copy()
                self.check_block(env_copy, body)

    def check_block(self, env: Env, block: BlockStatement):
        for stmt in block.statements:
            self.check_statement(env, stmt)
        return env

    def check_function(self, func: Function):
        env = self.check_args(func.args)
        self.check_block(env, func.body)

    def check(self, prog: Program):
        if len(prog.functions) != 1 or prog.functions[0].name != "main":
            self.error()
        self.check_function(prog.functions[0])

    def error(self):
        raise Exception("TypeChecker Error")